/*
 * (C) Copyright 2007-2012
 * Allwinner Technology Co., Ltd. <www.allwinnertech.com>
 * Benn Huang <benn@allwinnertech.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define pr_fmt(fmt) "sunxi: " fmt

#include <linux/module.h>
#include <linux/io.h>

#include <plat/platform.h>
#include <plat/system.h>

#define SC_CHIP_ID_REG	(SW_VA_SRAM_IO_BASE + 0x24)

#define SC_CHIP_ID_EN_MASK	0x1
#define SC_CHIP_ID_EN_OFF	15
#define SC_CHIP_ID_EN	(SC_CHIP_ID_EN_MASK<<SC_CHIP_ID_EN_OFF)

#define SC_CHIP_ID_MASK	0xffff
#define SC_CHIP_ID_OFF	16
#define SC_CHIP_ID		(SC_CHIP_ID_MASK<<SC_CHIP_ID_OFF)

/*
 * BROM
 */
struct brom_header {
	u32 jump_instruction;	/* one intruction jumping to real code */
	char magic[8];		/* ="eGON.BRM",  not C-style string. */
	u32 length;		/* generated by PC */
	char Boot_vsn[4];	/* Boot version */
	char eGON_vsn[4];	/* eGON version */
	char platform[8];	/* platform information */
};

static inline const struct brom_header *_sunxi_brom(void)
{
	const struct brom_header *brom = (void*)SW_VA_BROM_BASE + 0x4000;
	if (strncmp(brom->magic, "eGON.BRM", 8) == 0)
		return brom;
	else
		return NULL;
}

int sunxi_pr_brom(void)
{
	const struct brom_header *brom = _sunxi_brom();
	if (brom) {
		pr_info("BROM Ver: %.4s %.4s %.8s\n",
			brom->Boot_vsn,
			brom->eGON_vsn,
			brom->platform);
		return 1;
	}
	pr_err("BROM not found\n");
	return 0;
}

u32 sunxi_brom_chip_id(void)
{
	const struct brom_header *brom = _sunxi_brom();
	if (brom) {
		char *endptr;
		u32 chip_id = simple_strtoul(brom->platform, &endptr, 10);
		if (*endptr == '\0')
			return chip_id;

		pr_err("BROM: failed to parse chip-id: \"%s\"\n",
		       brom->platform);
	}
	return SUNXI_UNKNOWN_MACH;
}
EXPORT_SYMBOL(sunxi_brom_chip_id);

/*
 * SC
 */
u32 sunxi_sc_chip_id(void)
{
	u32 chip_id, reg_val;

	/* enable chip_id reading */
	reg_val = readl(SC_CHIP_ID_REG);
	writel(reg_val | SC_CHIP_ID_EN, SC_CHIP_ID_REG);

	reg_val = readl(SC_CHIP_ID_REG);
	chip_id = ((reg_val&SC_CHIP_ID)>>SC_CHIP_ID_OFF) & SC_CHIP_ID_MASK;

	switch (chip_id) {
	case 0x1623:
		return SUNXI_MACH_SUN4I;
	case 0x1625:
		return SUNXI_MACH_SUN5I;
	case 0x1633:
		return SUNXI_MACH_SUN6I;
	case 0x1651:
		return SUNXI_MACH_SUN7I;
	default:
		pr_err("SC: failed to identify chip-id 0x%04x (*0x%08x == 0x%08x)\n",
		       chip_id, SC_CHIP_ID_REG, reg_val);
		return SUNXI_UNKNOWN_MACH;
	}
}
EXPORT_SYMBOL(sunxi_sc_chip_id);

/*
 */
u32 sunxi_chip_id(void)
{
	static u32 chip_id;

	if (unlikely(chip_id == 0))
		chip_id = sunxi_sc_chip_id();

	return chip_id;
}
EXPORT_SYMBOL(sunxi_chip_id);

int sunxi_pr_chip_id(void)
{
	u32 chip_id = sunxi_chip_id();
	const char *soc_family, *name;
	int rev;

	if (sunxi_is_sun4i())
		soc_family = "sun4i";
	else if (sunxi_is_sun5i())
		soc_family = "sun5i";
	else if (sunxi_is_sun6i())
		soc_family = "sun6i";
	else if (sunxi_is_sun7i())
		soc_family = "sun7i";
	else
		soc_family = "sunNi?";

	if (sunxi_is_a10())
		name = "A10";
	else if (sunxi_is_a13())
		name = "A13";
	else if (sunxi_is_a12())
		name = "A12";
	else if (sunxi_is_a10s())
		name = "A10s";
	else if (sunxi_is_a31())
		name = "A31";
	else if (sunxi_is_a20())
		name = "A20";
	else
		name = NULL;

	rev = sunxi_soc_rev();
	if (rev)
		pr_info("Allwinner %s revision %c (AW%u/%s) detected.\n",
			name?name:"A??", 'A' + rev - 1, chip_id, soc_family);
	else
		pr_info("Allwinner %s (AW%u/%s) detected.\n",
			name?name:"A??", chip_id, soc_family);

	return name?1:0;
}

static inline void reg_dump(const char *name, u32 reg, unsigned len)
{
	unsigned i, j;

	for (i=0; i<len; ) {
		pr_info("soc-detect: %s (0x%08x):", name, reg);

		for (j=0; i<len && j<4; i++, j++, reg += 0x04) {
			u32 val = readl(reg);
			printk(" %08x", val);
		}

		printk("\n");
	}
}

enum sw_ic_ver sw_get_ic_ver(void)
{
	static enum sw_ic_ver ver;

	if (likely(ver))
		return ver;

	if (sunxi_is_sun4i()) {
		u32 val = readl(SW_VA_TIMERC_IO_BASE + 0x13c);
		val = (val >> 6) & 0x3;

		if (val == 0)
			ver = SUNXI_VER_A10A;
		else if (val == 3)
			ver = SUNXI_VER_A10B;
		else
			ver = SUNXI_VER_A10C;
	} else if (sunxi_is_sun5i()) {
		u32 val = readl(SW_VA_SID_IO_BASE + 0x08);
		val = (val >> 12) & 0xf;
		switch (val) {
		case 0:	ver = SUNXI_VER_A12; break;
		case 3: ver = SUNXI_VER_A13; break;
		case 7: ver = SUNXI_VER_A10S; break;
		default:
			goto unknown_chip;
		}

		val = readl(SW_VA_SID_IO_BASE+0x00);
		val = (val >> 8) & 0xffffff;

		if (val == 0 || val == 0x162541)
			ver += SUNXI_REV_A;
		else if (val == 0x162542)
			ver += SUNXI_REV_B;
		else {
			const char *name;
			if (ver == SUNXI_VER_A13)
				name = "A13";
			else if (ver == SUNXI_VER_A12)
				name = "A12";
			else
				name = "A10S";

			pr_err("unrecongnized %s revision (%x)\n",
			       name, val);

			reg_dump("SID", SW_VA_SID_IO_BASE, 4);
		}
	} else if (sunxi_is_sun6i())
		ver = SUNXI_VER_A31;
	else if (sunxi_is_sun7i())
		ver = SUNXI_VER_A20;

	goto done;

unknown_chip:
	pr_err("unrecognized IC (chip-id=%u)\n", sunxi_chip_id());
	ver = SUNXI_VER_UNKNOWN;

	if (sunxi_is_sun5i())
		reg_dump("SSE", SW_VA_SSE_IO_BASE, 1);
	reg_dump("SID", SW_VA_SID_IO_BASE, 4);
done:
	return ver;
}
EXPORT_SYMBOL(sw_get_ic_ver);

int sw_get_chip_id(struct sw_chip_id *chip_id)
{
	chip_id->sid_rkey0 = readl(SW_VA_SID_IO_BASE);
	chip_id->sid_rkey1 = readl(SW_VA_SID_IO_BASE+0x04);
	chip_id->sid_rkey2 = readl(SW_VA_SID_IO_BASE+0x08);
	chip_id->sid_rkey3 = readl(SW_VA_SID_IO_BASE+0x0C);

	return 0;
}
EXPORT_SYMBOL(sw_get_chip_id);
